/*
 * Copyright(C) 2010 Neil Jagdish Patel
 * Copyright(C) 2010 Canonical Ltd.
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * version 3.0 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3.0 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Authored by Neil Jagdish Patel <neil.patel@canonical.com>
 * Authored by Ken VaDine <ken.vandine@canonical.com>
 */

[DBus (name = "com.Gwibber.Streams")]
private interface StreamsInterface : Object {
        public abstract string Messages (string stream, string account, int time, string transient, string recipient, string orderby, string order, int limit) throws GLib.IOError;
        public abstract string Get (string id) throws GLib.IOError;
        public abstract string List () throws GLib.IOError;
        public abstract string Create (string query) throws GLib.IOError;
        public abstract void Delete (string id) throws GLib.IOError;
        public signal void Updated (string id);
        public signal void Created (string id);
        public signal void Deleted (string id);
}

namespace Gwibber
{
    public class Streams : Object
    {
        private const string streams_name  = "com.Gwibber.Streams";
        private const string streams_path  = "/com/gwibber/Streams";

        private StreamsInterface streams_service;
        private Gwibber.Service service;
	private Gwibber.Utils utils;
	private Gwibber.Messages messages_service;
	private Gwibber.Searches searches_service;
        private Dee.Model? _stream_model = null;
        private Dee.Model? _transient_model = null;
        private unowned Dee.ResourceManager resources;
        private Gee.HashMap <string, unowned Dee.ModelIter?> seen;

        public Dee.Model stream_model {
          get { 
            if (_stream_model != null) {
              return _stream_model;
            } else {
              _stream_model = streams_model (false);
              return _stream_model;
            }
          }
          set {
            if (_stream_model != value)
              _stream_model = value;
          }
        }

        public Dee.Model transient_model {
          get {
            if (_transient_model != null) {
              return _transient_model;
            } else {
              _transient_model = streams_model (true);
              return _transient_model;
            }
          }
          set {
            if (_transient_model != value)
              _transient_model = value;
          }
        }

	/** 
            Streams::is_available:
            @arg0: The current state

            Emitted when com.Gwibber.Streams availability state changes
	*/
	public signal void is_available(bool is_up);

        public abstract signal void updated (string id);
        public abstract signal void deleted (string id);
        public signal void created(string id);


    public Streams ()
    {
            try
            {
                streams_service = Bus.get_proxy_sync(BusType.SESSION,
                                                          streams_name,
                                                          streams_path);
                utils = new Gwibber.Utils();
                utils.setup(streams_name);
                utils.available.connect(streams_available);
                messages_service = new Gwibber.Messages ();
                searches_service = new Gwibber.Searches ();
                service = new Gwibber.Service ();
                streams_service.Updated.connect((source) => {
                    updated (source);
                });
                streams_service.Created.connect((source) => {
                    stream_created (source);
                });
                streams_service.Deleted.connect((source) => {
                    stream_deleted (source);
                });

            }
            catch (GLib.IOError e)
            {
                warning ("Unable to get Gwibber Streams: "+e.message);
            }
            resources = Dee.ResourceManager.get_default ();
            seen = new Gee.HashMap <string, unowned Dee.ModelIter?> ();
            var _m = resources.load ("gwibber.startup") as Dee.SequenceModel;
            if (!(_m is Dee.SequenceModel))
            {
                _m = new Dee.SequenceModel ();
                _m.set_schema ("s");
                _m.append ("start");
                resources.store (_m, "gwibber.startup");
            }
        }


        /**
         * com.Gwibber.Streams
         **/
        public string messages(string stream, string account, int time, string? transient)
        {
            if (transient == null)
              transient = "0";

            try
            {
                return streams_service.Messages(stream, account, time, transient, "0", "time", "DESC", 1000);
            }
            catch (GLib.IOError e)
            {
                warning (e.message);
                return "";
            }
          
        }   

        public Dee.Model? stream_filter_model (Dee.Model model, string stream)
        {
            Dee.Filter stream_filter;
            // If home stream requested, return full model
            if (stream == "home") {
              return model;
            } else {
              stream_filter = new Dee.Filter.for_key_column (StreamModelColumn.STREAM, stream);
            }
            var filtered_model = new Dee.FilterModel (stream_filter, model);
            return filtered_model;
        }

        public Dee.Model? transient_filter_model (Dee.Model model, string transient)
        {
            Dee.Filter transient_filter;
            transient_filter = new Dee.Filter.for_key_column (StreamModelColumn.TRANSIENT, transient);
            var filtered_model = new Dee.FilterModel (transient_filter, model);
            return filtered_model;
        }

        private Dee.SequenceModel? create_model()
        {
            var m = new Dee.SequenceModel();
            m.set_schema ("as", "s", "s", "s", "s", "b", "s", "s", "s", "s", "s", "s", "s", "s", "s", "s", "d", "s", "s", "s", "s", "s", "s", "s", "s", "s", "s", "s", "s", "s");
            return m;
        }

	public Dee.Model? streams_model(bool transients)
	{
            if (!transients)
            {
              debug ("Getting non-transient model");
              Dee.SequenceModel? model = resources.load ("gwibber.stream_model") as Dee.SequenceModel;
              debug ("stream_model from resources has %u rows", model.get_n_rows ());
              if (!(model is Dee.SequenceModel))
              {
                debug ("Didn't get model from resource manager, creating a new one");
                model = create_model ();
              }
              Timeout.add_seconds (5, () => {
                  model.clear ();
                  refresh_model ();
                  return false;
              });
              messages_service.message.connect((change, data) => {
                  on_refresh(change, data);});
              return model;
            }
            else
            {
              debug ("Getting transient model");
              Dee.SequenceModel? model = resources.load ("gwibber.transient_model") as Dee.SequenceModel;
              debug ("transient_model from resources has %u rows", model.get_n_rows ());
              if (!(model is Dee.SequenceModel))
              {
                debug ("Didn't get the transient model from resource manager, creating a new one");
                model = create_model ();
              }
              Timeout.add_seconds (5, () => {
                  model.clear ();
                  refresh_transient_model ();
                  return false;
              });
              searches_service.created.connect(() => {
                debug ("New search created");
                service.refresh ();
                service.loading_complete.connect(() => {
                  debug ("Got loading complete, refreshing model");
                  });
                });
              created.connect(() => {
                debug ("New stream created");
                service.refresh ();
                service.loading_complete.connect(() => {
                  debug ("Got loading complete, refreshing model");
                  });
                });
              return model;
            }
        }

        private void on_refresh(string change, string data)
        {
            //debug ("on_refresh %s - %s", change, data);
            var parser = new Json.Parser();
            parser.load_from_data(data, -1);
            unowned Json.Node node = parser.get_root();
            Json.Object obj = node.get_object();
            if (obj != null)
              parse_message (obj, change);
        }

        private void refresh_model()
        {
            debug ("refresh_model");
            var streams = new List<string> ();
            streams.append("replies");
            streams.append("private");
            streams.append("public");
            streams.append("images");
            streams.append("links");
            streams.append("videos");
            streams.append("messages");
            foreach (string stream in streams)
            {
              string msgs = messages(stream, "all", 0, null);
              parse_messages (msgs);
            }
            debug ("Storing gwibber.stream_model");
            resources.store ((Dee.SequenceModel)stream_model, "gwibber.stream_model");
            debug ("stream_model has %u rows", stream_model.get_n_rows ());
        }

        private void refresh_transient_model()
        {
            debug ("refresh_transient_model");
            var searches = searches_service.list ();
            foreach (Gee.HashMap <string?, string?> x in searches)
            {
              string msgs = messages("all", "all", 0, x["id"]);
              parse_messages (msgs);
            }
            var streams_list = list ();
            foreach (Gee.HashMap <string?, string?> x in streams_list)
            {
              string msgs = messages("all", x["account"], 0, x["id"]);
              parse_messages (msgs);
            }
            debug ("Storing gwibber.transient_model");
            resources.store ((Dee.SequenceModel)transient_model, "gwibber.transient_model");
            debug ("transient_model has %u rows", transient_model.get_n_rows ());
        }
       
        private void parse_messages (string data)
        {
            var parser = new Json.Parser();
            parser.load_from_data(data, -1);
            Json.Array nodeArray = parser.get_root().get_array();

            for(int i = 0; i < nodeArray.get_length(); i++) {
              Json.Object obj = nodeArray.get_element(i).get_object();
              parse_message (obj, "new");
            }
        }

        private void parse_message (Json.Object obj, string change)
        {
            string transient = obj.get_string_member("transient");
            
            if (transient == null)
              transient = "0";
            var _model = stream_model;
            if (transient.length > 1)
              _model = transient_model;

            if (!(_model is Dee.Model))
            {
              debug ("ERROR: Not a valid model, transient is %s", transient);
              return;
            }
            if (transient.length > 1 && change != "new")
              return;

            unowned Dee.ModelIter? iter = null;
            var members = obj.get_members();
            string _account = obj.get_string_member("account");
            //debug ("_account is %s", _account);
            string _stream = obj.get_string_member("stream");
            string _service = obj.get_string_member("service");
            string _operation = obj.get_string_member("operation");
            if (_operation == "send")
              return;
            string _mid = obj.get_string_member("id");
            string _acct = _account + ":" + _service + ":" + _mid;
            var builder = new VariantBuilder (new VariantType ("as"));
            builder.add ("s", _acct);
            Json.Object _sender_obj = obj.get_object_member("sender");
            string _sender = _sender_obj.get_string_member("name");
            string _sender_nick = null;
            if (_sender_obj.has_member("nick"))
              _sender_nick = _sender_obj.get_string_member("nick");

            bool _from_me = _sender_obj.get_boolean_member("is_me");
            uint _time = (uint)obj.get_double_member("time");
            if (_time < 1)
            {
                _time = (uint)obj.get_int_member("time");
            }
            string _html = obj.get_string_member("content");
            string _text = obj.get_string_member("text");

            string _textid = "";
            foreach (var x in _text.replace(".", "").split(" "))
              if ((!x.has_prefix("http://")) && (!x.has_prefix("https://")) 
                && (!x.has_prefix("gwibber:/")) && (!x.has_prefix("#"))
                && (!x.has_prefix("!")))
                  _textid += x + " ";

            _textid = _textid.strip();
            // FIXME: the transient == 0 disables dupe checking for transients, we should fix that
            //string _icon = yield service.avatar_path(_sender_obj.get_string_member("image"));
            string _icon = _sender_obj.get_string_member("image");
            string _url = obj.get_string_member("url");

            string _source = null;
            string _reply_nick = null;
            string _reply_name = null;
            string _reply_url = null;
            double _likes = 0;
            string _retweeted_by_nick = null;
            string _retweeted_by_name = null;
            string _retweeted_by_id = null;
            string _link_picture = null;
            string _link_name = null;
            string _link_url = null;
            string _link_description = null;
            string _link_caption = null;
            string _link_icon = null;
            string _image_url = null;
            string _image_src = null;
            string _image_thumb = null;
            string _image_name = null;

            foreach (string member in members)
            {
              if (member == "retweeted_by")
              {
                Json.Object _retweet_obj = obj.get_object_member(member);
                _retweeted_by_nick = _retweet_obj.get_string_member("nick");
                _retweeted_by_name = _retweet_obj.get_string_member("name");
                _retweeted_by_id = _retweet_obj.get_int_member("id").to_string();
              }

              if (member == "source")
              {
                _source = obj.get_string_member("source");
                if (_source != null)
                  _source = _source.replace("rel=\"nofollow\"", "");
              }
              if (member == "reply")
              {
                Json.Object _reply_obj = obj.get_object_member("reply");
                _reply_nick = _reply_obj.get_string_member("nick");
                _reply_name = _reply_obj.get_string_member("name");
                _reply_url = GLib.Markup.escape_text (_reply_obj.get_string_member("url"));
              }
              if (member == "likes")
              {
                Json.Object _likes_obj = obj.get_object_member("likes");
                _likes = (double)_likes_obj.get_int_member("count");
              }
              if (member == "favorited")
              {
                if (obj.get_boolean_member("favorited"))
                {
                  _likes = -1;
                }
              }

              if (member == "link")
              {
                Json.Object _link_obj = obj.get_object_member(member);
                _link_picture = _link_obj.get_string_member("picture");
                _link_name = _link_obj.get_string_member("name");
                _link_url = GLib.Markup.escape_text (_link_obj.get_string_member("url"), -1);
                _link_description = _link_obj.get_string_member("description");
                _link_caption = _link_obj.get_string_member("caption");
                _link_icon = _link_obj.get_string_member("icon");
              }
              if (member == "photo")
              {
                Json.Object _photo_obj = obj.get_object_member(member);
                _image_url = GLib.Markup.escape_text (_photo_obj.get_string_member("url"));
                _image_src = GLib.Markup.escape_text (_photo_obj.get_string_member("picture"));
                _image_name = _photo_obj.get_string_member("name");
              }
              if (member == "images")
              {
                Json.Array _images_array = obj.get_array_member(member);
                Json.Object _image_obj = _images_array.get_element(0).get_object();
                _image_url = GLib.Markup.escape_text (_image_obj.get_string_member("url"));
                _image_src = GLib.Markup.escape_text (_image_obj.get_string_member("src"));
                _image_thumb = GLib.Markup.escape_text (_image_obj.get_string_member("thumb"));
                _image_name = _image_obj.get_string_member("name");
              }
            }
            /* FIXME: hacky scrubbing of the html, we should find a 
               better way */
            _html = _html.replace("&", "&amp;");
            //_html = _html.replace(" &amp; ", " & ");
            _html = _html.replace("class=\"nick\"", "");
            _html = _html.replace("class=\"hash\"", "");

            string _t = utils.generate_time_string(_time);

            if (seen.has_key(_textid) && transient == "0" && _stream != "replies")
            {
              iter = seen.get(_textid);
              if (iter == null)
                return;
              try
              {
                string[] _accounts_array = (string[])_model.get_value (iter, StreamModelColumn.ACCOUNTS);
                if (!(_acct in _accounts_array))
                {
                  foreach (var a in _accounts_array)
                    builder.add ("s", a);
                  var _accounts = new Variant ("as",  builder);
                  _model.set_value (iter, StreamModelColumn.ACCOUNTS, _accounts);
                }
                _model.set_value (iter, StreamModelColumn.MESSAGE, _text);
                _model.set_value (iter, StreamModelColumn.HTML, _html);
                _model.set_value (iter, StreamModelColumn.TIMESTRING, _t);
                _model.set_value (iter, StreamModelColumn.LIKES, _likes);
                iter = null;
              } catch {
                debug ("Failed to update record for mid: %s", _mid);
              }
              return;
            }

            var _accounts = new Variant ("as",  builder);

            //debug ("row is: %s %s %s %s %s", _stream, _sender_nick, _time.to_string(), _source, transient);
            
            iter = _model.append(
              _accounts,
              _stream,
              transient,
              _sender,
              _sender_nick,
              _from_me,
              _time.to_string(),
              _text,
              _html,
              _icon,
              _url,
              _source,
              _t,
              _reply_nick,
              _reply_name,
              _reply_url,
              _likes,
              _retweeted_by_nick,
              _retweeted_by_name,
              _retweeted_by_id,
              _link_picture,
              _link_name,
              _link_url,
              _link_description,
              _link_caption,
              _link_icon,
              _image_url,
              _image_src,
              _image_thumb,
              _image_name
              );

            if (_textid.length > 1 && iter != null)
              seen[_textid] = iter;
            iter = null;
            //debug ("_model has %u ROWS", _model.get_n_rows ());
        }


        /**
         * com.Gwibber.Streams
         **/
        public Gee.HashMap <string?, string?> lookup(string id)
        {
            try
            {
                var result = streams_service.Get(id);
                var parser = new Json.Parser();
                parser.load_from_data(result, -1);
                Json.Object obj = parser.get_root().get_object();
                var members = obj.get_members();
                var stream_map = new Gee.HashMap <string?, string?> ();
                foreach (string member in members)
                {
                  stream_map[member] = obj.get_string_member(member);
                }
                return stream_map;
            }
            catch (GLib.IOError e)
            {
                warning (e.message);
                return null;
            }

        }


        /**
         * com.Gwibber.Streams
         **/
        public void delete(string id)
        {
            try
            {
                streams_service.Delete(id);
            }
            catch (GLib.IOError e)
            {
                warning (e.message);
            }
            return;
        }


        /**
         * com.Gwibber.Streams
         **/

        public string create(string account, string name)
        {
            try
            {
                return streams_service.Create("{\"account\": \"" + account + "\", \"parameters\": {\"count\": 50, \"id\": \"" + name + "\"}, \"name\": \"" + name + "\", \"operation\": \"user_messages\"}");
            }
            catch (GLib.IOError e)
            {
                warning (e.message);
                return "";
            }
        }


        /**
         * com.Gwibber.Streams
         **/
        public List<Gee.HashMap?> list()
        {
            try
            {
                var streams_list = new List<Gee.HashMap<string?, string?>> ();
                var result = streams_service.List();
                var parser = new Json.Parser();
                parser.load_from_data(result, -1);
                Json.Array nodeArray = parser.get_root().get_array();
                for(int i = 0; i < nodeArray.get_length(); i++) {
                  var stream_map = new Gee.HashMap <string?, string?> ();
                  var members = nodeArray.get_element(i).get_object().get_members();
                  foreach (string member in members)
                  {
                    stream_map[member] = nodeArray.get_element(i).get_object().get_string_member(member);
                  }
                  streams_list.append(stream_map);
                }
                return streams_list;
            }
            catch (GLib.IOError e)
            {
                warning (e.message);
                return null;
            }

        }


        public void stream_deleted(string data)
        {
            try {
                var parser = new Json.Parser();
                parser.load_from_data(data, -1);
                var root_object = parser.get_root().get_object();
                string id = root_object.get_string_member("id");
                deleted(id);
            } catch(GLib.Error e) {
                warning (e.message);
            }
        }

        public void stream_created(string data)
        {
            try {
                var parser = new Json.Parser();
                parser.load_from_data(data, -1);
                var root_object = parser.get_root().get_object();
                string id = root_object.get_string_member("id");
                created(id);
            } catch(GLib.Error e) {
                warning (e.message);
            }
        }

	public void streams_available(bool is_up)
	{
		is_available(is_up);
	}
    }
}
